/*
 * System Interface Library for games
 * Copyright (c) 2007-2016 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/posix/ioqueue.c: Asynchronous I/O functions.
 */

#define IN_SYSDEP

#include "src/base.h"
#include "src/memory.h"
#include "src/sysdep.h"
#include "src/sysdep/posix/ioqueue.h"

#ifdef SIL_INCLUDE_TESTS
/* Include this first so we can properly hide the time() declaration to
 * avoid a shadowing warning. */
# define time __UNUSED_time
# include <time.h>
# undef time
#endif

#include <errno.h>
#include <fcntl.h>
#include <pthread.h>
#include <sys/types.h>
#include <unistd.h>

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* Structure holding data for an I/O request. */
typedef struct IORequest IORequest;
struct IORequest {
    int index;          // Array index (constant for each element).
    int id;             // Request ID (currently always equal to index+1).
    int next_pending;   // Array index of next pending request.
    uint8_t in_use;     // True if this entry is in use.
    uint8_t complete;   // True if this request has completed.
    uint8_t cancelled;  // True if this request was cancelled.

    /* Type of request and request details. */
    enum {IOQ_OPEN, IOQ_READ} type;
    union {
        struct {
            char *path;  // Locally-allocated copy.
            int flags;
        } open;
        struct {
            int fd;
            void *buf;
            ssize_t count;
            off_t pos;
        } read;
    };

    /* Request result (valid only when .complete is true). */
    ssize_t result;
    /* Error code (errno) from request.  Only valid if the request failed. */
    int error;

    /* Condition variable used to signal completion of the request.
     * Note that we allocate this separately because the requests[]
     * array may be moved by the mem_realloc() call in get_new_request(),
     * and some systems (at least Linux) misbehave if a pthread_cond_t
     * value is moved while a thread is waiting on it. */
    pthread_cond_t *completion_event;
};

/* Array of IORequest structures, dynamically resized as necessary. */
static IORequest *requests;
static int requests_size;

/* Index of last used entry in requests[], or -1 if no entries are used. */
static int requests_last_used;

/* Index of first and last pending requests in the queue.  If the queue is
 * empty, both are set to -1. */
static int first_pending = -1, last_pending = -1;

/* Mutex for accessing the requests array and queue pointers.  The mutex
 * must be held as long a pointer to any array element is in use (since
 * the array may be reallocated by another thread at any time). */
static pthread_mutex_t requests_mutex = PTHREAD_MUTEX_INITIALIZER;

/* Thread ID for background I/O thread, or 0 if the thread has not yet
 * been started. */
static pthread_t io_thread_id = 0;

/* Condition variable used to signal that a new requests has been enqueued. */
static pthread_cond_t enqueue_event = PTHREAD_COND_INITIALIZER;

/* Flag set by ioq_reset() to stop the background I/O thread. */
static uint8_t thread_stop_flag;

/*-----------------------------------------------------------------------*/

#ifdef SIL_INCLUDE_TESTS

/* Flags set by test control routines. */
static uint8_t move_on_realloc = 0;
static uint8_t block_io_thread = 0;
static uint8_t unblock_on_wait = 0;
static uint8_t permfail_next_read = 0;
static uint8_t tempfail_next_read = 0;
static uint8_t iofail_next_read = 0;

#endif

/*-----------------------------------------------------------------------*/

/* Local routine declarations. */

/**
 * get_new_request:  Return the index of a currently unused request in the
 * request array.
 *
 * On entry, requests_mutex is assumed to be unlocked.  On successful
 * return, requests_mutex is locked.
 *
 * [Return value]
 *     requests[] array index of request, or -1 on error.
 */
static int get_new_request(void);

/**
 * get_request_by_id:  Return the index of the request with the given ID.
 *
 * On entry, requests_mutex is assumed to be unlocked.  On successful
 * return, requests_mutex is locked.
 *
 * [Parameters]
 *     id: Request ID (must be an in-use request block).
 * [Return value]
 *     requests[] array index of request, or -1 on error or invalid ID.
 */
static int get_request_by_id(int id);

/**
 * release_request:  Release (mark unused) the request with the given index.
 *
 * On entry, requests_mutex is assumed to be locked.
 *
 * [Parameters]
 *     index: requests[] index of request to release.
 */
static void release_request(int index);

/**
 * resize_requests:  Resize the requests[] array to the given size.
 *
 * On entry, requests_mutex is assumed to be locked.
 *
 * [Parameters]
 *     new_size: New size (number of entries) for requests[] array.
 */
static int resize_requests(int new_size);

/**
 * enqueue_request:  Enqueue the given request for background processing,
 * starting the background processing thread if necessary.  If the thread
 * cannot be started, the request is processed synchronously.
 *
 * On entry, requests_mutex is assumed to be locked.
 *
 * [Parameters]
 *     index: requests[] index of request to enqueue.
 */
static void enqueue_request(int index);

/**
 * dequeue_request:  Dequeue the first queued request and return its index.
 *
 * On entry, requests_mutex is assumed to be locked.
 *
 * [Return value]
 *     requests[] index of dequeued request, or -1 if no requests were queued.
 */
static int dequeue_request(void);

/**
 * process_request:  Process the given request.
 *
 * On entry, requests_mutex is assumed to be locked.
 *
 * [Parameters]
 *     index: requests[] index of request to process.
 */
static void process_request(int index);

/**
 * io_thread:  Background thread which loops until thread_stop_flag becomes
 * true, dequeueing and performing I/O requests as they are queued.
 *
 * [Parameters]
 *     thread_start_event: Condition variable used to signal that the
 *         thread has started and locked requests_mutex.
 * [Return value]
 *     NULL
 */
static void *io_thread(void *);

/**
 * start_io_thread:  Start the I/O thread running.
 *
 * On entry, requests_mutex is assumed to be locked.
 *
 * [Return value]
 *     True on success, false on error.
 */
static int start_io_thread(void);

/*************************************************************************/
/************************** Interface routines ***************************/
/*************************************************************************/

int ioq_open(const char *path, int flags)
{
    if (!path || (flags & O_ACCMODE) != O_RDONLY) {
        errno = EINVAL;
        return 0;
    }

    char * const path_copy = mem_strdup(path, MEM_ALLOC_TEMP);
    if (UNLIKELY(!path_copy)) {
        DLOG("Failed to copy path %s", path);
        errno = ENOMEM;
        return 0;
    }

    const int index = get_new_request();
    if (UNLIKELY(index < 0)) {
        DLOG("Failed to get a request block");
        mem_free(path_copy);
        errno = ENOMEM;
        return 0;
    }

    requests[index].type       = IOQ_OPEN;
    requests[index].open.path  = path_copy;
    requests[index].open.flags = flags;
    enqueue_request(index);

    const int id = requests[index].id;
    pthread_mutex_unlock(&requests_mutex);
    return id;
}

/*-----------------------------------------------------------------------*/

int ioq_read(int fd, void *buf, ssize_t count, off_t pos)
{
    if (fd == -1 || buf == NULL || count < 0 || pos < 0) {
        errno = EINVAL;
        return 0;
    }

#ifdef SIL_INCLUDE_TESTS
    if (permfail_next_read) {
        DLOG("Force-failing request with ENOMEM");
        errno = ENOMEM;
        permfail_next_read = 0;
        return 0;
    }
    if (tempfail_next_read) {
        DLOG("Force-failing request with EAGAIN");
        errno = EAGAIN;
        tempfail_next_read = 0;
        return 0;
    }
#endif

    const int index = get_new_request();
    if (UNLIKELY(index < 0)) {
        DLOG("Failed to get a request block");
        errno = ENOMEM;
        return 0;
    }

    requests[index].type       = IOQ_READ;
    requests[index].read.fd    = fd;
    requests[index].read.buf   = buf;
    requests[index].read.count = count;
    requests[index].read.pos   = pos;
    enqueue_request(index);

    const int id = requests[index].id;
    pthread_mutex_unlock(&requests_mutex);
    return id;
}

/*-----------------------------------------------------------------------*/

int ioq_poll(int request)
{
    const int index = get_request_by_id(request);
    if (UNLIKELY(index < 0)) {
        DLOG("Invalid request ID: %d", request);
        errno = ESRCH;
        return 1;
    }

    const int complete = requests[index].complete;

    pthread_mutex_unlock(&requests_mutex);
    return complete;
}

/*-----------------------------------------------------------------------*/

ssize_t ioq_wait(int request, int *error_ret)
{
#ifdef SIL_INCLUDE_TESTS
    if (unblock_on_wait) {
        block_io_thread = 0;
        BARRIER();
    }
#endif

    const int index = get_request_by_id(request);
    if (UNLIKELY(index < 0)) {
        DLOG("Invalid request ID: %d", request);
        if (error_ret) {
            *error_ret = ESRCH;
        }
        errno = ESRCH;
        return -1;
    }

    while (!requests[index].complete) {
        pthread_cond_wait(requests[index].completion_event, &requests_mutex);
    }

    if (requests[index].type == IOQ_OPEN) {
        mem_free(requests[index].open.path);
        requests[index].open.path = NULL;
    }

    const ssize_t result = requests[index].result;
    if (error_ret) {
        *error_ret = requests[index].error;
    }
    release_request(index);

    pthread_mutex_unlock(&requests_mutex);
    return result;
}

/*-----------------------------------------------------------------------*/

void ioq_cancel(int request)
{
    const int index = get_request_by_id(request);
    if (UNLIKELY(index < 0)) {
        DLOG("Invalid request ID: %d", request);
        errno = ESRCH;
        return;
    }

    if (requests[index].complete) {
        if (requests[index].type == IOQ_OPEN && requests[index].result != -1) {
            close(requests[index].result);
        }
    } else {
        /* Skip over this request when it's dequeued. */
        requests[index].cancelled = 1;
    }
    requests[index].result = -1;
    requests[index].error = ECANCELED;

    pthread_mutex_unlock(&requests_mutex);
}

/*-----------------------------------------------------------------------*/

void ioq_reset(void)
{
    if (io_thread_id) {
        pthread_mutex_lock(&requests_mutex);
        thread_stop_flag = 1;
        pthread_cond_signal(&enqueue_event);
        pthread_mutex_unlock(&requests_mutex);
        pthread_join(io_thread_id, NULL);
        thread_stop_flag = 0;
        io_thread_id = 0;
        pthread_cond_destroy(&enqueue_event);
        pthread_cond_init(&enqueue_event, NULL);
        pthread_mutex_destroy(&requests_mutex);
        pthread_mutex_init(&requests_mutex, NULL);
    }

    resize_requests(0);
    first_pending = last_pending = -1;
}

/*************************************************************************/
/************************* Test control routines *************************/
/*************************************************************************/

#ifdef SIL_INCLUDE_TESTS

/*-----------------------------------------------------------------------*/

void TEST_posix_ioqueue_move_on_realloc(int enable)
{
    move_on_realloc = (enable != 0);
}

/*-----------------------------------------------------------------------*/

void TEST_posix_ioqueue_block_io_thread(int enable)
{
    BARRIER();
    block_io_thread = (enable != 0);
    BARRIER();
}

/*-----------------------------------------------------------------------*/

void TEST_posix_ioqueue_unblock_on_wait(int enable)
{
    unblock_on_wait = (enable != 0);
}

/*-----------------------------------------------------------------------*/

void TEST_posix_ioqueue_permfail_next_read(int enable)
{
    permfail_next_read = (enable != 0);
}

/*-----------------------------------------------------------------------*/

void TEST_posix_ioqueue_tempfail_next_read(int enable)
{
    tempfail_next_read = (enable != 0);
}

/*-----------------------------------------------------------------------*/

void TEST_posix_ioqueue_iofail_next_read(int enable)
{
    iofail_next_read = (enable != 0);
}

/*-----------------------------------------------------------------------*/

#endif  // SIL_INCLUDE_TESTS

/*************************************************************************/
/**************************** Local routines *****************************/
/*************************************************************************/

static int get_new_request(void)
{
    pthread_mutex_lock(&requests_mutex);

    int index;
    for (index = 0; index < requests_size; index++) {
        if (!requests[index].in_use) {
            break;
        }
    }
    if (index >= requests_size) {
        if (!resize_requests(index + 1)) {
            pthread_mutex_unlock(&requests_mutex);
            return -1;
        }
    }
    if (index > requests_last_used) {
        requests_last_used = index;
    }

    requests[index].in_use = 1;
    requests[index].complete = 0;
    requests[index].cancelled = 0;
    requests[index].error = 0;
    return index;
}

/*-----------------------------------------------------------------------*/

static int get_request_by_id(int id)
{
    if (id < 1 || id > requests_size) {
        return -1;
    }

    pthread_mutex_lock(&requests_mutex);

    const int index = id - 1;
    if (!requests[index].in_use) {
        DLOG("Request ID %d is valid but not in use", id);
        pthread_mutex_unlock(&requests_mutex);
        return -1;
    }

    return index;
}

/*-----------------------------------------------------------------------*/

static void release_request(int index)
{
    PRECOND(index >= 0 && index < requests_size, return);

    requests[index].in_use = 0;
    if (index == requests_last_used) {
        for (; index >= 0; index--) {
            if (requests[index].in_use) {
                break;
            }
        }
        requests_last_used = index;
        const int new_size = index + 1;
        /* Shrink the array if a significant portion of it is unused, but
         * always leave a few entries allocated so we're not repeatedly
         * allocating and freeing for solitary requests. */
        if (new_size <= requests_size/2 - 5) {
            resize_requests(new_size);
        }
    }
}

/*-----------------------------------------------------------------------*/

static int resize_requests(int new_size)
{
    PRECOND(new_size >= 0);

    /* Free resources associated with request blocks that are about to be
     * deallocated (when shrinking or freeing the array). */
    for (int index = new_size; index < requests_size; index++) {
        if (requests[index].in_use) {
            if (requests[index].type == IOQ_OPEN) {
                if (requests[index].complete && requests[index].result != -1) {
                    close(requests[index].result);
                }
                mem_free(requests[index].open.path);
            }
        }
        pthread_cond_destroy(requests[index].completion_event);
        mem_free(requests[index].completion_event);
    }

    /* If reallocating to zero length, just free the array and return. */
    if (new_size == 0) {
        mem_free(requests);
        requests = NULL;
        requests_size = 0;
        return 1;
    }

    /* Reallocate the array to its new size.  If reallocation fails when
     * shrinking the array, we proceed with the original pointer and
     * pretend the shrink succeeded. */
    IORequest *new_requests =
#ifdef SIL_INCLUDE_TESTS
        move_on_realloc ? mem_alloc(sizeof(*requests) * new_size, 0, 0) :
#endif
        mem_realloc(requests, sizeof(*requests) * new_size, 0);
    if (!new_requests) {
        if (UNLIKELY(new_size > requests_size)) {
            DLOG("Failed to expand requests array to %d entries", new_size);
            return 0;
        } else {
            new_requests = requests;
        }
    }
#ifdef SIL_INCLUDE_TESTS
    else if (move_on_realloc) {
        memcpy(new_requests, requests,
               sizeof(*requests) * min(requests_size, new_size));
        mem_free(requests);
    }
#endif
    requests = new_requests;
    const int old_size = requests_size;
    requests_size = new_size;

    /* Initialize and allocate resources for new array entries (when
     * expanding the array). */
    for (int i = old_size; i < requests_size; i++) {
        requests[i].index = i;
        requests[i].id = requests[i].index + 1;
        requests[i].next_pending = -1;
        requests[i].in_use = 0;
        requests[i].completion_event =
            mem_alloc(sizeof(*requests[i].completion_event), 0, 0);
        if (UNLIKELY(!requests[i].completion_event)) {
            DLOG("Failed to allocate completion condvar for index %d", i);
            requests_size = i;
            if (requests_size == 0) {
                mem_free(requests);
                requests = NULL;
            } else {
                new_requests = mem_realloc(
                    requests, sizeof(*requests) * requests_size, 0);
                if (new_requests) {
                    requests = new_requests;
                }
            }
            return 0;
        }
        pthread_cond_init(requests[i].completion_event, NULL);
    }

    return 1;
}

/*-----------------------------------------------------------------------*/

static void enqueue_request(int index)
{
    PRECOND(index >= 0 && index < requests_size, return);

#ifdef SIL_INCLUDE_TESTS
    if (requests[index].type == IOQ_READ) {
        if (iofail_next_read) {
            requests[index].result = -1;
            requests[index].error = EIO;
            requests[index].complete = 1;
            pthread_cond_signal(requests[index].completion_event);
            iofail_next_read = 0;
            return;
        }
    }
#endif

    if (!io_thread_id) {
        if (UNLIKELY(!start_io_thread())) {
            process_request(index);
            requests[index].complete = 1;
            pthread_cond_signal(requests[index].completion_event);
            return;
        }
    }

    if (last_pending < 0) {
        first_pending = index;
    } else {
        requests[last_pending].next_pending = index;
    }
    last_pending = index;
    requests[index].next_pending = -1;

    pthread_cond_signal(&enqueue_event);
}

/*-----------------------------------------------------------------------*/

static int dequeue_request(void)
{
    int dequeued = first_pending;
    if (dequeued >= 0) {
        if (requests[dequeued].next_pending < 0) {
            first_pending = last_pending = -1;
        } else {
            first_pending = requests[dequeued].next_pending;
        }
        requests[dequeued].next_pending = -1;
    }
    return dequeued;
}

/*-----------------------------------------------------------------------*/

static void process_request(int index)
{
    PRECOND(index >= 0 && index < requests_size, return);
    IORequest * const request = &requests[index];

    switch (request->type) {

      case IOQ_OPEN:
        request->result = open(request->open.path, request->open.flags, 0666);
        break;

      case IOQ_READ:
        request->result = pread(request->read.fd, request->read.buf,
                                request->read.count, request->read.pos);
        break;

    }  // switch (request->type)

    request->error = errno;
}

/*************************************************************************/
/************************* Background I/O thread *************************/
/*************************************************************************/

static void *io_thread(void *thread_start_event)
{
    pthread_mutex_lock(&requests_mutex);
    pthread_cond_signal(thread_start_event);
    while (!thread_stop_flag) {
#ifdef SIL_INCLUDE_TESTS
        BARRIER();
        while (block_io_thread) {
            pthread_mutex_unlock(&requests_mutex);
            nanosleep(&(struct timespec){0, 1000000}, NULL);  // 1 msec
            pthread_mutex_lock(&requests_mutex);
        }
#endif
        const int index = dequeue_request();
        if (index < 0) {
            pthread_cond_wait(&enqueue_event, &requests_mutex);
        } else {
            if (!requests[index].cancelled) {
                process_request(index);
            }
            requests[index].complete = 1;
            pthread_cond_signal(requests[index].completion_event);
        }
    }
    pthread_mutex_unlock(&requests_mutex);
    return NULL;
}

/*-----------------------------------------------------------------------*/

static int start_io_thread(void)
{
    pthread_cond_t thread_start_event = PTHREAD_COND_INITIALIZER;
    const int error = pthread_create(&io_thread_id, NULL,
                                     io_thread, &thread_start_event);
    if (UNLIKELY(error != 0)) {
        io_thread_id = 0;
        static int warned = 0;
        if (!warned) {
            DLOG("Failed to create background I/O thread: %s",
                 strerror(error));
            warned = 1;
        }
        return 0;
    }
    pthread_cond_wait(&thread_start_event, &requests_mutex);
    pthread_cond_destroy(&thread_start_event);
    return 1;
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
