/*
 * System Interface Library for games
 * Copyright (c) 2007-2016 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/windows/condvar.c: System-level condition variable routines
 * for Windows.
 */

/*
 * This implementation is based on "Strategies for Implementing POSIX
 * Condition Variables on Win32" by Douglas C. Schmidt and Irfan Pyarali
 * (http://www.cs.wustl.edu/~schmidt/win32-cv-1.html).  That paper asserts
 * that the method is fair in part because SignalObjectAndWait()'s
 * atomicity ensures that all threads have a chance to respond to a signal
 * before any other thread can lock the mutex; unfortunately, it turns out
 * that SignalObjectAndWait() is not in fact atomic (making it a singularly
 * useless function), so this implementation has the potential for unfair
 * scheduling behavior.
 *
 * Windows Vista adds native condition variables, but they can't be used
 * with mutex objects, so they're useless to us.
 */

#define IN_SYSDEP

#include "src/base.h"
#include "src/math.h"
#include "src/memory.h"
#include "src/sysdep.h"
#include "src/sysdep/windows/internal.h"

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* Data structure for a condition variable. */
typedef struct SysCondVar SysCondVar;
struct SysCondVar {
    /* Number of threads waiting on this condition variable. */
    int num_waiters;
    /* Lock protecting access to num_waiters. */
    CRITICAL_SECTION num_waiters_lock;
    /* Semaphore used to wake waiters. */
    HANDLE wait_sem;
    /* Event object used to signal when all threads have received a
     * broadcast event. */
    HANDLE waiters_done;
    /* Flag: Was the last signal operation a broadcast? */
    uint8_t was_broadcast;
};

/*************************************************************************/
/************************** Interface routines ***************************/
/*************************************************************************/

SysCondVarID sys_condvar_create(void)
{
    SysCondVar *condvar = mem_alloc(sizeof(*condvar), 0, 0);
    if (!condvar) {
        DLOG("No memory for condition variable");
        goto error_return;
    }

    condvar->wait_sem = CreateSemaphore(NULL, 0, 0x7FFFFFFF, NULL);
    if (!condvar->wait_sem) {
        DLOG("Failed to create semaphore for condition variable: %s",
             windows_strerror(GetLastError()));
        goto error_free_condvar;
    }

    condvar->waiters_done = CreateEvent(NULL, FALSE, FALSE, NULL);
    if (!condvar->waiters_done) {
        DLOG("Failed to create event object for condition variable: %s",
             windows_strerror(GetLastError()));
        goto error_destroy_wait_sem;
    }

    condvar->num_waiters = 0;
    InitializeCriticalSection(&condvar->num_waiters_lock);
    condvar->was_broadcast = 0;
    return (SysCondVarID)condvar;

  error_destroy_wait_sem:
    CloseHandle(condvar->wait_sem);
  error_free_condvar:
    mem_free(condvar);
  error_return:
    return 0;
}

/*-----------------------------------------------------------------------*/

void sys_condvar_destroy(SysCondVarID condvar_)
{
    SysCondVar *condvar = (SysCondVar *)condvar_;
    CloseHandle(condvar->waiters_done);
    CloseHandle(condvar->wait_sem);
    mem_free(condvar);
}

/*-----------------------------------------------------------------------*/

int sys_condvar_wait(SysCondVarID condvar_, SysMutexID mutex_, float timeout)
{
    SysCondVar *condvar = (SysCondVar *)condvar_;
    HANDLE mutex = (HANDLE)mutex_;

    /* Add this thread to the set of waiters for broadcasts. */
    EnterCriticalSection(&condvar->num_waiters_lock);
    condvar->num_waiters++;
    LeaveCriticalSection(&condvar->num_waiters_lock);

    /* Normally, a condition variable should release the mutex and enter a
     * wait state as a single atomic operation.  Windows doesn't have such
     * an atomic operation (SignalObjectAndWait() looks like it should be
     * that operation but in fact is not atomic), so it's possible for
     * another thread to squeeze between these two calls, take the mutex,
     * and signal the condition variable.  However, since we use a counting
     * semaphore rather than a boolean event flag for the wait operation,
     * there's no loss of correctness; the WaitForSingleObject() call will
     * just return immediately instead of waiting, and since we increment
     * the waiter count before unlocking the mutex, broadcast operations
     * will always include this thread in the semaphore release count. */
    ReleaseMutex(mutex);
    const int retval =
        (WaitForSingleObject(condvar->wait_sem, timeout_to_ms(timeout))
         == WAIT_OBJECT_0);

    /* Remove this thread from the waiter set.  Also check whether we
     * need to signal completion to an in-progress broadcast operation. */
    EnterCriticalSection(&condvar->num_waiters_lock);
    condvar->num_waiters--;
    const int last_waiter = condvar->was_broadcast && condvar->num_waiters==0;
    LeaveCriticalSection(&condvar->num_waiters_lock);

    /* If this thread was the last one to wake from a broadcast operation,
     * signal that the broadcast is complete.  To guarantee fairness of
     * scheduling, this would need to be an atomic operation that both
     * signalled the waiters_done event and waited on the caller's mutex
     * (thus ensuring that this thread gets onto the mutex wait list before
     * any other waiter has a chance to resume), but unfortunately we can't
     * do that in Windows. */
    if (last_waiter) {
        SetEvent(condvar->waiters_done);
    }

    /* Relock the caller's mutex before returning. */
    WaitForSingleObject(mutex, INFINITE);
    return retval;
}

/*-----------------------------------------------------------------------*/

void sys_condvar_signal(SysCondVarID condvar_, int broadcast)
{
    SysCondVar *condvar = (SysCondVar *)condvar_;

    EnterCriticalSection(&condvar->num_waiters_lock);
    const int num_waiters = condvar->num_waiters;
    if (!num_waiters) {
        /* No threads waiting, so nothing to do. */
        LeaveCriticalSection(&condvar->num_waiters_lock);
        return;
    }

    if (broadcast) {
        condvar->was_broadcast = 1;
        /* The critical section is still held here, so no other threads
         * can enter sys_condvar_wait() until after the ReleaseSemaphore()
         * call.  (And in any case, the condition variable API requires
         * the mutex to be held for signal and broadcast as well as wait
         * operations, so no thread can call sys_condvar_wait() while this
         * function is executing anyway.) */
        ReleaseSemaphore(condvar->wait_sem, num_waiters, NULL);
        LeaveCriticalSection(&condvar->num_waiters_lock);
        WaitForSingleObject(condvar->waiters_done, INFINITE);
        /* The semaphore value may still be nonzero here, if any threads
         * timed out on the semaphore wait but we read num_waiters before
         * the timing-out thread decremented it, so clear out any leftover
         * value.  We rely on the API requirement to hold the mutex when
         * calling this function in order to ensure correctness. */
        for (int i = 0; i < num_waiters; i++) {
            if (WaitForSingleObject(condvar->wait_sem, 0) != WAIT_OBJECT_0) {
                break;
            }
        }
        /* This is safe without locking the critical section because of
         * the API requirement to hold the mutex. */
        condvar->was_broadcast = 0;
    } else {
        LeaveCriticalSection(&condvar->num_waiters_lock);
        ReleaseSemaphore(condvar->wait_sem, 1, NULL);
    }
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
